from subprocess import Popen, PIPE, STDOUT
from argparse import Namespace
import copy
import json
import sys
import os
import traceback
import fasteners
from datetime import datetime

from . import cache, filesys
from .cache import Cache
from .utils import merge_dicts, substract_dict_keys

DEFAULT_ARGS_KEYS = ['exp_config', 'exp_dir', 'exp_is_complete', 'exp_force', 'exp_no_wait', 'exp_hash']
DEFAULT_CONFIG_KEYS = ['executable', 'exp_time']

def init_args(executable):
    args = {
        'executable': executable,
        'exp_time': datetime.now().strftime("%Y-%m-%d_%H-%M-%S")
    }
    return args

def setup(*args, **kwargs):
    return Setup(*args, **kwargs)

class Setup:
    def __init__(self, parser, hash_ignore=[]):
        parser.add_argument("--exp-config", default='{}', type=json.loads)
        parser.add_argument("--exp-dir", default=False, action="store_const", const=True)
        parser.add_argument("--exp-is-complete", default=False, action="store_const", const=True)
        parser.add_argument("--exp-force", default=False, action="store_const", const=True)
        parser.add_argument("--exp-no-wait", default=False, action="store_const", const=True)
        parser.add_argument("--exp-hash", default=False, action="store_const", const=True)

        self.parser = parser

        self._cache = Cache()
        self._hash_ignore = hash_ignore
        self._run_lock = None
    
    def __enter__(self):
        executable = filesys.relative_root_path(sys.argv[0])

        args = init_args(executable)
        args = merge_dicts(args, dict(vars(self.parser.parse_args())))
        args = merge_dicts(args, args['exp_config'])

        self._all_args = args

        user_args = substract_dict_keys(args, DEFAULT_ARGS_KEYS + DEFAULT_CONFIG_KEYS)
        config_args = substract_dict_keys(args, DEFAULT_ARGS_KEYS)
        hash_args = substract_dict_keys(args, 
            [k for k in DEFAULT_CONFIG_KEYS if k not in ['executable']] +
            DEFAULT_ARGS_KEYS + self._hash_ignore
        )
        
        self.args = Namespace(**user_args)
        

        exists = self._cache.exists(hash_args)
        self.dir = self._cache.get_dir(hash_args) if exists else self._cache.assign_dir(hash_args)

        os.makedirs(self.dir, exist_ok=True)

        path = os.path.join(self.dir, 'config.json')
        if not os.path.exists(path) or args['exp_force']:
            with open(path, 'w') as out_file:
                json.dump(config_args, out_file, indent=4)

        if args['exp_hash']:
            print(cache.get_hash(hash_args))
            exit(0)

        if args['exp_dir']:
            print(self.dir)
            exit(0)
        
        if args['exp_is_complete']:
            print(self._cache.is_complete(hash_args))
            exit(0)

        self._run_lock = fasteners.InterProcessLock(os.path.join(self.dir, '.run.lock'))
        self._run_lock.acquire()

        err_filename = os.path.join(self.dir, 'error.log')
        if os.path.exists(err_filename):
            os.remove(err_filename)
        
        if self._cache.is_complete(hash_args) and not args['exp_force']:
            print('*** Using cached data on {}'.format(self.dir))
            self._run_lock.release()
            exit(0)
        
        return self
    
    
    def __exit__(self, exc_type, exc_value, tb):
        if exc_type is not None:
            tb_message = ''.join(traceback.format_exception(exc_type, exc_value, tb))
            err_filename = os.path.join(self.dir, 'error.log')
            with open(err_filename, 'w') as err_file:
                err_file.write(tb_message)
                
            self._run_lock.release()
            return False
        
        hash_args = substract_dict_keys(self._all_args, 
            [k for k in DEFAULT_CONFIG_KEYS if k not in ['executable']] +
            DEFAULT_ARGS_KEYS + self._hash_ignore
        )

        self._cache.set_complete(hash_args)

        self._run_lock.release()

        return True



class Experiment:
    def __init__(self, executable, req_args, command):
        self.executable = filesys.relative_root_path(executable)
        self.command = command
        self.args = req_args
        
        self._cache = Cache()
        self._last_full_hash = None
        self._last_local_hash = cache.get_hash(self.args)

        dir = self.get_dir()
        
        path = os.path.join(dir, 'config.json')
        with open(path, 'r') as in_file:
            self.args = json.load(in_file)

    def run(self, custom_command=None, use_cached=True, wait=True):
        tmp_args = init_args(self.executable)
        tmp_args = merge_dicts(tmp_args, self.args)

        command = custom_command if custom_command != None else self.command
        command = command.format(**tmp_args)
        command_parts = command.split(' ')
        if not use_cached:
            command_parts.append('--exp-force')
        if not wait:
            command_parts.append('--exp-no-wait')
        command_parts += ["--exp-config", "{}".format(json.dumps(self.args))]
        
        exe = Popen(command_parts, stdout=PIPE, stderr=PIPE)
        out, err = exe.communicate()

    def get_hash(self):
        curr_local_hash = cache.get_hash(self.args)
        if curr_local_hash == self._last_local_hash and self._last_full_hash is not None:
            return self._last_full_hash
        self._last_local_hash = curr_local_hash

        if self._cache.exists(self.args):
            return curr_local_hash
        
        tmp_args = init_args(self.executable)
        tmp_args = merge_dicts(tmp_args, self.args)

        command = self.command.format(**tmp_args)
        command_parts = command.split(' ')
        command_parts.append("--exp-hash")
        command_parts += ["--exp-config", "{}".format(json.dumps(self.args))]

        exe = Popen(command_parts, stdout=PIPE, stderr=PIPE)
        out, err = exe.communicate()

        lines = out.decode(sys.stdin.encoding).split('\n')
        err_msg = err.decode(sys.stdin.encoding)

        if len(lines) < 2:
            if len(err) > 0:
                raise Exception(err_msg)
            else:
                raise Exception("error: Command did not print an output.")

        hash = lines[-2]

        if not self._cache.exists(hash):
            if len(err) > 0:
                raise Exception(err_msg)
            else:
                raise Exception("error: Command returned invalid hash.")

        self._last_full_hash = hash
        self._cache.merge_hashes(curr_local_hash, hash)

        return hash

    def get_dir(self):
        return self._cache.get_dir(self.get_hash())

    def is_complete(self):
        return self._cache.is_complete(self.get_hash())


import sys
import os

from . import filesys

MAIN_USAGE_MESSAGE = """
usage: xlab command ...

Options:

positional arguments:
  command
    project
"""

def project(args):
    if len(args) != 1:
        print("error: Invalid arguments.")
        exit()
    
    if args[0] == 'init':
        root = os.getcwd()
        
        dirs = filesys.Directories()
        dirs.set_root(root)


def main():
    if len(sys.argv) <= 1:
        print(MAIN_USAGE_MESSAGE)
        exit()
    
    command = sys.argv[1]
    args = sys.argv[2:]

    if command == 'project':
        exe = project
    else:
        print("error: No command 'xlab {}'.".format(command))
        exit()
    
    exe(args)


import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
	long_description = fh.read()

requires_list = [
    'fasteners'
]

setuptools.setup(
	name="xlab",
	version="0.0.16",
	author="Cesar Salcedo",
	author_email="cesar.salcedo@utec.edu.pe",
	description="Experiment execution tool for automation in research-oriented projects.",
	long_description=long_description,
	long_description_content_type="text/markdown",
	url="https://github.com/csalcedo001/xlab",
	packages=setuptools.find_packages(),
	classifiers=[
		"Programming Language :: Python :: 3",
                "License :: OSI Approved :: Apache Software License",
		"Operating System :: OS Independent",
	],
	python_requires='>=3.6',
	install_requires=requires_list,
	entry_points={
		'console_scripts':[
			'xlab=xlab.cli:main'
		]
	}
)



import os
import sys
import json
import pickle
import fasteners

_dirs = {}

def find_root_dir():
    try:
        filename = os.path.realpath(sys.argv[0])
        dirname = os.path.dirname(filename) if os.path.isfile(filename) else filename

        curr_dir = dirname
    except:
        curr_dir = os.getcwd()
    
    abs_root = os.path.abspath(os.sep)

    while curr_dir != abs_root and '.exp' not in os.listdir(curr_dir):
        curr_dir = os.path.dirname(curr_dir)
    
    if '.exp' not in os.listdir(curr_dir):
        curr_dir = None
    
    return curr_dir

def relative_root_path(path):
    path = os.path.realpath(path)
    root = dirs.root()

    if root[-len(os.sep):] != os.sep:
        root = root + os.sep
    
    if len(path) < len(root) or path[:len(root)] != root:
        raise Exception("error: Path does not belong to project. Received {} which was expected to be found within {}.".format(path, root))

    return path[len(root):]


class Directories:
    def __init__(self):
        root = find_root_dir()

        self._init_q = root != None
        if self._init_q:
            self.set_root(root)
    
    def set_root(self, root):
        exp_path = os.path.join(root, '.exp')
        os.makedirs(exp_path, exist_ok=True)

        _dirs['root'] = root
        _dirs['exp'] = exp_path

        self._init_q = True
    
    def root(self):
        if not self._init_q:
            print("error: Could not find '.exp' folder. Try running 'xlab project init' on your project root directory.")
            exit(1)
        return _dirs['root']

    def exp_path(self):
        if not self._init_q:
            print("error: Could not find '.exp' folder. Try running 'xlab project init' on your project root directory.")
            exit(1)
        return _dirs['exp']

    def runs_path(self):
        if 'runs' not in _dirs:
            root = self.root()

            runs_path = os.path.join(root, 'runs')
            os.makedirs(runs_path, exist_ok=True)

            _dirs['runs'] = runs_path

        return _dirs['runs']

dirs = Directories()



class MetadataLoader:
    def __init__(self, path, name):
        filename = '{}.json'.format(name)
        self.path = os.path.join(path, filename)

        lock_filename = '.{}.lock'.format(name)
        lock_path = os.path.join(path, lock_filename)
        self.lock = fasteners.InterProcessReaderWriterLock(lock_path)

        self.lock.acquire_write_lock()
        if not os.path.exists(self.path):
            with open(self.path, 'w') as out_file:
                json.dump({
                    'next_id': 0
                }, out_file)
        self.lock.release_write_lock()
    
    def next_id(self):
        self.lock.acquire_write_lock()
        with open(self.path, 'r') as in_file:
            metadata = json.load(in_file)

        id = metadata['next_id']
        metadata['next_id'] += 1

        with open(self.path, 'w') as out_file:
            json.dump(metadata, out_file)
        self.lock.release_write_lock()

        return id



class HashmapLoader:
    def __init__(self, path, name):
        filename = '{}.pkl'.format(name)
        self.path = os.path.join(path, filename)

        lock_filename = '.{}.lock'.format(name)
        lock_path = os.path.join(path, lock_filename)
        self.lock = fasteners.InterProcessReaderWriterLock(lock_path)
        self.locked_q = False

        self.lock.acquire_write_lock()
        if not os.path.exists(self.path):
            with open(self.path, 'wb') as out_file:
                pickle.dump({}, out_file)
        self.lock.release_write_lock()
    
    def load(self):
        assert not self.locked_q

        self.lock.acquire_read_lock()
        with open(self.path, 'rb') as in_file:
            hashmap = pickle.load(in_file)
        self.lock.release_read_lock()

        return hashmap
    
    def load_and_lock_acquire(self):
        assert not self.locked_q
        self.locked_q = True

        self.lock.acquire_write_lock()
        with open(self.path, 'rb') as in_file:
            hashmap = pickle.load(in_file)
        
        return hashmap
    
    def save_and_lock_release(self, hashmap):
        assert self.locked_q
        self.locked_q = False

        with open(self.path, 'wb') as out_file:
            pickle.dump(hashmap, out_file)
        self.lock.release_write_lock()



from argparse import Namespace
import copy
import sys
import os

def merge_dicts(a, b, output_type=None):
    a_type = type(a)
    b_type = type(b)

    if a_type == Namespace:
        a = dict(vars(a))
    if b_type == Namespace:
        b = dict(vars(b))

    a = copy.deepcopy(a)

    for key in b:
        val = b[key]
        if type(val) == dict and key in a and type(a[key]) == dict:
            a[key] = merge_dicts(a[key], val)
        else:
            a[key] = val

    if output_type == None:
        output_type = a_type
    if output_type == Namespace:
        a = Namespace(**a)

    return a

def substract_dict_keys(a, keys):
    a = copy.deepcopy(a)
    for key in keys:
        if key in a:
            del a[key]
    
    return a


import json
import hashlib
import copy
import os

from . import filesys

def sort_args(args):
    args = copy.deepcopy(args)
    if type(args) == dict:
        for key in args:
            args[key] = sort_args(args[key])
        return sorted(args.items())
    if type(args) == list:
        return [sort_args(x) for x in args]

    return args

# Cache functions
def get_args_hash(args):
    hash = hashlib.sha224(json.dumps(sort_args(args), separators=(',', ':')).encode('utf-8')).hexdigest()
    return hash

def get_hash(args_or_hash):
    if type(args_or_hash) == dict:
        hash = get_args_hash(args_or_hash)
    elif type(args_or_hash) == str:
        hash = args_or_hash
    
    return hash

# Cache class
class Cache:
    def __init__(self):
        self.metadata_loader = filesys.MetadataLoader(filesys.dirs.exp_path(), 'metadata')
        self.hashmap_loader = filesys.HashmapLoader(filesys.dirs.exp_path(), 'hashmap')

    def exists(self, args_or_hash):
        hash = get_hash(args_or_hash)
        hashmap = self.hashmap_loader.load()

        return hash in hashmap

    def is_complete(self, args_or_hash):
        hash = get_hash(args_or_hash)
        hashmap = self.hashmap_loader.load()

        return hash in hashmap and hashmap[hash][1]

    def get_dir(self, args_or_hash):
        hash = get_hash(args_or_hash)
        hashmap = self.hashmap_loader.load()

        if hash in hashmap:
            return hashmap[hash][0]
        else:
            raise Exception('error: Hash not found in cache.')

    def assign_dir(self, args):
        id = self.metadata_loader.next_id()
        path = os.path.join(filesys.dirs.runs_path(), str(id))
        hash = get_hash(args)

        hashmap = self.hashmap_loader.load_and_lock_acquire()
        hashmap[hash] = [path, False]
        self.hashmap_loader.save_and_lock_release(hashmap)
        
        return path

    def set_complete(self, args_or_hash):
        hash = get_hash(args_or_hash)

        hashmap = self.hashmap_loader.load_and_lock_acquire()
        hashmap[hash][1] = True
        self.hashmap_loader.save_and_lock_release(hashmap)

    def merge_hashes(self, new_hash, old_hash):
        hashmap = self.hashmap_loader.load_and_lock_acquire()
        hashmap[new_hash] = hashmap[old_hash]
        self.hashmap_loader.save_and_lock_release(hashmap)
